@@ -38,11 +38,14 @@ from abc import ABC
 from copy import copy
 from datetime import datetime, timedelta
 from enum import Enum
+from glob import glob
+import json
 from math import exp, inf, sqrt
+import os
 from os import PathLike
 from pathlib import Path
 from random import gauss, random
-from typing import Union
+from typing import Mapping, Union
 
 from adcircpy.forcing.winds.best_track import convert_value, FileDeck, Mode, VortexForcing
 from dateutil.parser import parse as parse_date
@@ -55,7 +58,9 @@ from pyproj import CRS, Transformer
 from pyproj.enums import TransformDirection
 from shapely.geometry import LineString
 
-from ensembleperturbation.utilities import units
+from ensembleperturbation.utilities import get_logger, units
+
+LOGGER = get_logger('perturbation.atcf')
 
 AIR_DENSITY = 1.15 * units.kilogram / units.meters ** 3
 
@@ -188,7 +193,11 @@ class VortexPerturbedVariable(ABC):
         self.__default = default
 
     def perturb(
-        self, vortex_dataframe: DataFrame, values: [float], times: [datetime],
+        self,
+        vortex_dataframe: DataFrame,
+        values: [float],
+        times: [datetime],
+        inplace: bool = False,
     ) -> DataFrame:
         """
         perturb the variable within physical bounds
@@ -196,9 +205,14 @@ class VortexPerturbedVariable(ABC):
         :param vortex_dataframe: ATCF dataframe containing track info
         :param values: values for each forecast time (VT)
         :param times: forecast times (VT)
+        :param inplace: modify dataframe in-place
         :return: updated ATCF dataframe with perturbed values
         """
 
+        if not inplace:
+            # make a deepcopy to preserve the original dataframe
+            vortex_dataframe = vortex_dataframe.copy(deep=True)
+
         all_values = vortex_dataframe[self.name].values + values
         vortex_dataframe[self.name] = [
             min(self.upper_bound, max(value, self.lower_bound)).magnitude
@@ -478,7 +492,11 @@ class CrossTrack(VortexPerturbedVariable):
         )
 
     def perturb(
-        self, vortex_dataframe: DataFrame, values: [float], times: [datetime],
+        self,
+        vortex_dataframe: DataFrame,
+        values: [float],
+        times: [datetime],
+        inplace: bool = False,
     ) -> DataFrame:
         """
         offset_track(df_,VT,cross_track_errors)
@@ -487,9 +505,14 @@ class CrossTrack(VortexPerturbedVariable):
         :param vortex_dataframe: ATCF dataframe containing track info
         :param values: cross-track errors [nm] for each forecast time (VT)
         :param times: forecast times (VT)
+        :param inplace: modify dataframe in-place
         :return: updated ATCF dataframe with different longitude latitude locations based on perpendicular offsets set by the cross_track_errors
         """
 
+        if not inplace:
+            # make a deepcopy to preserve the original dataframe
+            vortex_dataframe = vortex_dataframe.copy(deep=True)
+
         # Get the coordinates of the track
         track_coords = vortex_dataframe[['longitude', 'latitude']].values.tolist()
 
@@ -566,12 +589,7 @@ class CrossTrack(VortexPerturbedVariable):
                 )
             )
 
-        degree = PintType(units.degree)
         vortex_dataframe['latitude'], vortex_dataframe['longitude'] = zip(*new_coordinates)
-        vortex_dataframe['longitude'] = vortex_dataframe['longitude'].astype(
-            degree, copy=False
-        )
-        vortex_dataframe['latitude'] = vortex_dataframe['latitude'].astype(degree, copy=False)
 
         return vortex_dataframe
 
@@ -637,7 +655,11 @@ class AlongTrack(VortexPerturbedVariable):
         )
 
     def perturb(
-        self, vortex_dataframe: DataFrame, values: [float], times: [datetime],
+        self,
+        vortex_dataframe: DataFrame,
+        values: [float],
+        times: [datetime],
+        inplace: bool = False,
     ) -> DataFrame:
         """
         interpolate_along_track(df_,VT,along_track_errros)
@@ -646,9 +668,14 @@ class AlongTrack(VortexPerturbedVariable):
         :param vortex_dataframe: ATCF dataframe containing track info
         :param values: along-track errors for each forecast time (VT)
         :param times: forecast timed (VT)
+        :param inplace: modify dataframe in-place
         :return: updated ATCF dataframe with different longitude latitude locations based on interpolated errors along track
         """
 
+        if not inplace:
+            # make a deepcopy to preserve the original dataframe
+            vortex_dataframe = vortex_dataframe.copy(deep=True)
+
         max_interpolated_points = 5  # maximum number of pts along line for each interpolation
 
         # Get the coordinates of the track
@@ -730,16 +757,21 @@ class AlongTrack(VortexPerturbedVariable):
                 )
             )
 
-        degree = PintType(units.degree)
         vortex_dataframe['latitude'], vortex_dataframe['longitude'] = zip(*new_coordinates)
-        vortex_dataframe['longitude'] = vortex_dataframe['longitude'].astype(
-            degree, copy=False
-        )
-        vortex_dataframe['latitude'] = vortex_dataframe['latitude'].astype(degree, copy=False)
 
         return vortex_dataframe
 
 
+VARIABLES = [
+    CentralPressure,
+    BackgroundPressure,
+    MaximumSustainedWindSpeed,
+    RadiusOfMaximumWinds,
+    CrossTrack,
+    AlongTrack,
+]
+
+
 class VortexPerturber:
     def __init__(
         self,
@@ -865,12 +897,14 @@ class VortexPerturber:
         variables: [VortexPerturbedVariable],
         directory: PathLike = None,
         alphas: [float] = None,
+        overwrite: bool = False,
     ) -> [Path]:
         """
         :param number_of_perturbations: number of perturbations to create
         :param variables: list of variable names, any combination of `["max_sustained_wind_speed", "radius_of_maximum_winds", "along_track", "cross_track"]`
         :param directory: directory to which to write
         :param alphas: list of floats meant to represent a point on the standard Gaussian distribution (see random.gauss function) for all variables except for "radius_of_maximum_winds" where should be list of floats in range [0, 1) (see random.random function). These alpha values are used to multiply error for perturbation; leave None for random
+        :param overwrite: overwrite existing files
         :returns: written filenames
         """
 
@@ -888,6 +922,11 @@ class VortexPerturber:
         for index, variable in enumerate(variables):
             if isinstance(variable, type):
                 variables[index] = variable()
+            elif isinstance(variable, str):
+                for existing_variable in VARIABLES:
+                    if variable == existing_variable.name:
+                        variables[index] = existing_variable()
+                        break
 
         if directory is None:
             directory = Path.cwd()
@@ -896,6 +935,10 @@ class VortexPerturber:
         if not directory.exists():
             directory.mkdir(parents=True, exist_ok=True)
 
+        if overwrite:
+            for filename in directory.iterdir():
+                os.remove(filename)
+
         # write out original fort.22
         self.forcing.write(directory / 'original.22', overwrite=True)
 
@@ -921,72 +964,103 @@ class VortexPerturber:
             copy=False,
         )
 
+        existing_filenames = glob(str(directory / 'vortex_*_variable_perturbation_*.22'))
+        if len(existing_filenames) > 0:
+            existing_filenames = sorted(existing_filenames)
+            last_index = int(existing_filenames[-1][-4]) + 1
+        else:
+            last_index = 1
+
         # for each variable, perturb the values and write each to a new `fort.22`
         output_filenames = []
-        for variable in variables:
-            print(f'writing perturbations for "{variable.name}"')
-            # Make the random pertubations based on the historical forecast errors
-            # Interpolate from the given VT to the storm_VT
-            if isinstance(variable, RadiusOfMaximumWinds):
-                storm_classification = storm_size
-            else:
-                storm_classification = storm_strength
+        for perturbation_number in range(last_index, number_of_perturbations + last_index):
+            LOGGER.info(
+                f'building perturbation {perturbation_number - last_index + 1} of {number_of_perturbations}'
+            )
 
-            historical_forecast_errors = variable.historical_forecast_errors[
-                storm_classification
-            ]
-            for column in historical_forecast_errors:
-                if isinstance(historical_forecast_errors[column].dtype, PintType):
-                    historical_forecast_errors[column] = historical_forecast_errors[
-                        column
-                    ].pint.magnitude
+            perturbed_data = original_data.copy(deep=True)
 
-            xp = historical_forecast_errors.index
-            fp = historical_forecast_errors.values
-            base_errors = [
-                interp(self.validation_times / timedelta(hours=1), xp, fp[:, ncol])
-                for ncol in range(len(fp[0]))
-            ]
+            # setting the alpha to the value from the input list
+            perturbation_alphas = alphas[perturbation_number - last_index]
 
-            for perturbation_index in range(1, number_of_perturbations + 1):
-                # make a deepcopy to preserve the original dataframe
-                perturbed_data = original_data.copy(deep=True)
-                perturbed_data = perturbed_data.astype(
-                    {
-                        variable.name: PintType(variable.unit)
-                        for variable in variables
-                        if variable.name in original_data
-                    },
-                    copy=False,
-                )
+            if perturbation_alphas is None:
+                perturbation_alphas = {variable.name: None for variable in variables}
+            elif not isinstance(perturbation_alphas, Mapping):
+                perturbation_alphas = {
+                    variable.name: perturbation_alphas for variable in variables
+                }
+
+            perturbation_alphas = {
+                variable.name
+                if isinstance(variable, type) and issubclass(variable, VortexPerturbedVariable)
+                else variable: alpha
+                for variable, alpha in perturbation_alphas.items()
+            }
+
+            for variable in variables:
+                alpha = perturbation_alphas[variable.name]
+
+                # Make the random pertubations based on the historical forecast errors
+                # Interpolate from the given VT to the storm_VT
+                if isinstance(variable, RadiusOfMaximumWinds):
+                    storm_classification = storm_size
+                else:
+                    storm_classification = storm_strength
+
+                historical_forecast_errors = variable.historical_forecast_errors[
+                    storm_classification
+                ]
+                for column in historical_forecast_errors:
+                    if isinstance(historical_forecast_errors[column].dtype, PintType):
+                        historical_forecast_errors[column] = historical_forecast_errors[
+                            column
+                        ].pint.magnitude
+
+                xp = historical_forecast_errors.index
+                fp = historical_forecast_errors.values
+                base_errors = [
+                    interp(self.validation_times / timedelta(hours=1), xp, fp[:, ncol])
+                    for ncol in range(len(fp[0]))
+                ]
 
-                # setting the alpha to the value from the input list
-                alpha = alphas[perturbation_index - 1]
                 # get the random perturbation sample
                 if variable.perturbation_type == PerturbationType.GAUSSIAN:
                     if alpha is None:
                         alpha = gauss(0, 1) / 0.7979
+                        perturbation_alphas[variable.name] = alpha
 
-                    print(f'Random gaussian variable = {alpha}')
+                    LOGGER.debug(f'gaussian alpha = {alpha}')
                     perturbation = base_errors[0] * alpha
                     if variable.unit is not None and variable.unit != units.dimensionless:
                         perturbation *= variable.unit
 
                     # add the error to the variable with bounds to some physical constraints
                     perturbed_data = variable.perturb(
-                        perturbed_data, values=perturbation, times=self.validation_times,
+                        perturbed_data,
+                        values=perturbation,
+                        times=self.validation_times,
+                        inplace=True,
                     )
                 elif variable.perturbation_type == PerturbationType.LINEAR:
                     if alpha is None:
                         alpha = random()
-                    print(f'Random number in [0,1) = {alpha}')
+                        perturbation_alphas[variable.name] = alpha
+
+                    LOGGER.debug(f'linear alpha [0,1) = {alpha}')
                     perturbation = -(base_errors[0] * (1.0 - alpha) + base_errors[1] * alpha)
                     if variable.unit is not None and variable.unit != units.dimensionless:
                         perturbation *= variable.unit
 
                     # subtract the error from the variable with physical constraint bounds
                     perturbed_data = variable.perturb(
-                        perturbed_data, values=perturbation, times=self.validation_times,
+                        perturbed_data,
+                        values=perturbation,
+                        times=self.validation_times,
+                        inplace=True,
+                    )
+                else:
+                    raise NotImplementedError(
+                        f'perturbation type "{variable.perturbation_type}" is not recognized'
                     )
 
                 if isinstance(variable, MaximumSustainedWindSpeed):
@@ -995,21 +1069,22 @@ class VortexPerturber:
                         perturbed_data
                     )
 
-                # remove units from data frame
-                for column in perturbed_data:
-                    if isinstance(perturbed_data[column].dtype, PintType):
-                        perturbed_data[column] = perturbed_data[column].pint.magnitude
+            # remove units from data frame
+            for column in perturbed_data:
+                if isinstance(perturbed_data[column].dtype, PintType):
+                    perturbed_data[column] = perturbed_data[column].pint.magnitude
 
-                # reset the dataframe
-                perturbed_forcing = copy(self.forcing)
-                perturbed_forcing.dataframe = perturbed_data
-
-                # write out the modified fort.22
-                output_filename = directory / f'{variable.name}_{perturbation_index}.22'
-                perturbed_forcing.write(
-                    output_filename, overwrite=True,
-                )
-                output_filenames.append(output_filename)
+            # write out the modified `fort.22`
+            perturbation_name = (
+                f'vortex_{len(variables)}_variable_perturbation_{perturbation_number}'
+            )
+            output_filename = directory / f'{perturbation_name}.22'
+            perturbed_forcing = copy(self.forcing)
+            perturbed_forcing.dataframe.loc[perturbed_data.index] = perturbed_data
+            perturbed_forcing.write(output_filename, overwrite=True)
+            output_filenames.append(output_filename)
+            with open(directory / f'{perturbation_name}.json', 'w') as output_json:
+                json.dump(perturbation_alphas, output_json, indent=2)
 
         return output_filenames
 
